<?php
/**
 * Plugin Name: WP Jsttyplay
 * Plugin URI: http://ba.sozo.ac.jp/myama/home/develop/jsttyplay/
 * Description: jsttyplay plugin for WordPress
 * Version: 0.1
 * Author: Michiru Yamaguchi
 * Author URI: http://ba.sozo.ac.jp/myama/
 *
 * Copyright (c) 2013 Michiru Yamaguchi
 *
 * As permitted by the BSD License, this work has been retrieved 
 * from https://github.com/encryptio/jsttyplay and republished here.
 *
 * WITHOUT ANY WARRANTY
 *
 * shortcode usage: [jsttyplay url="tty_file_path"]
 *  url : required
 *
 */
?>
<?php
add_action('admin_menu', 'jsttyplay_add_pages');
add_option('jsttyplay_font', 'fixed-7x13', '', 'yes');

function jsttyplay_add_pages() {
	add_options_page('WP Jsttyplay Options', 'Jsttyplay', 'manage_options', 'jsttyplayoptions', 'jsttyplay_options_page');
}

function jsttyplay_options_page() {
?>
<div class="wrap">
	<h2>WP Jsttyplay Options</h2>

	<form method="post" action="options.php">
		<?php wp_nonce_field('update-options'); ?>

		<table class="form-table">
			<tr valign="top">
				<th scope="row">Fonts</th>
				<td>
<?php
	$fontlist = array(
		'fixed-4x6', 'fixed-5x7', 'fixed-5x8', 'fixed-6x10', 
		'fixed-6x12', 'fixed-6x13', 'fixed-6x13B', 'fixed-6x13O', 
		'fixed-6x9', 'fixed-7x13', 'fixed-7x13B', 'fixed-7x13O', 
		'fixed-7x14', 'fixed-7x14B', 'fixed-8x13', 'fixed-8x13B', 
		'fixed-8x13O', 'fixed-9x15', 'fixed-9x15B', 'fixed-9x18', 
		'fixed-9x18B', 'fixed-10x20', 'fixed-12x13ja', 'fixed-18x18ja', 
		'fixed-18x18ko', 'profont-10', 'profont-11', 'profont-12', 
		'profont-15', 'profont-17', 'profont-22', 'profont-29', 
		'qemu-vgafont'
	);
	$font = get_option('jsttyplay_font');
?>
				<select name="jsttyplay_font">
<?php
	foreach ($fontlist as $f) {
		print('<option value="'.$f.'" ');
		if (strpos($font, $f) !== FALSE) {
			print('selected ');
		}
		print('>'.$f.'</option>');
	}
?>
				</select>
			</td>
		</tr>
	</table>

	<input type="hidden" name="action" value="update" />
	<input type="hidden" name="page_options" value="jsttyplay_font" />

	<p class="submit">
	<input type="submit" class="button-primary" value="<?php _e('Save Changes') ?>" />
	</p>
</form>
</div>

<?php
}


function allow_upload_tty( $mimes ) {
	$mimes['tty'] = 'image/x-ttyplay'; 
	return $mimes;
}
add_filter( 'upload_mimes', 'allow_upload_tty' );



add_shortcode('jsttyplay', 'jsttyplay_handler');

function jsttyplay_handler($atts, $content = null) {

	extract( shortcode_atts( array(
		'url' => null,
	), $atts ) );

	if (!isset($url)) {
		return 'require .tty file path [jsttyplay url=\'tty_file_path\']';
	}

	$content = get_jsttyplay_code($url);

	return $content;
}

function get_jsttyplay_code($url) {
	$plugin_base = plugins_url('', __FILE__);
	$font = get_option('jsttyplay_font');

/*
	wp_enqueue_script('vt-parse', get_template_directory_uri().'/js/vt/parse.js');
	wp_enqueue_script('vt-emulate', get_template_directory_uri().'/js/vt/emulate.js');
	wp_enqueue_script('ttyrec-browser', get_template_directory_uri().'/js/ttyrec-browser.js');
	wp_enqueue_script('binary-read', get_template_directory_uri().'/js/binary-read.js');
	wp_enqueue_script('vt-font', get_template_directory_uri().'/js/vt/font.js');
	wp_enqueue_script('vt-canvasview', get_template_directory_uri().'/js/vt/canvasview.js');
*/

$content = <<< EOM
<script src="{$plugin_base}/js/vt/parse.js"></script>
<script src="{$plugin_base}/js/vt/emulate.js"></script>
<script src="{$plugin_base}/js/ttyrec-browser.js"></script>
<script src="{$plugin_base}/js/binary-read.js"></script>
<script src="{$plugin_base}/js/vt/font.js"></script>
<script src="{$plugin_base}/js/vt/canvasview.js"></script>
<script>
	var tty_data;
	var nextFrameIdx = 0;

	var accurateTimeInterval = 1000/60; // max 60fps
	var frameJumpMax = 20;

	var viewReady = false;
	var playReady = false;

	var initialState;
	var vtview;
	setTimeout(function(){
		vtview = new VTCanvasView(document.getElementById('canvas'), {
			onReady: function () {
				viewReady = true;
				readyCheck();
			},
			//fontName: 'fixed-9x15'
			fontName: '{$font}'
		});
		initialState = { vtview: vtview.freeze(), nextFrameIdx: 0 };
	}, 20);

	get_binary_data_async("{$url}", function (data, err) {
		if ( err ) throw err;
			tty_data = TTYRecParse(data);

		playReady = true;
		readyCheck();
	});

	function readyCheck() {
		if ( viewReady && playReady )
			go();
	}

	var time_diff;

	function go() {
		var now = (new Date()).getTime()/1000;
		time_diff = now - tty_data[nextFrameIdx].time;

		nextFrame();
	}

	var nextFrameTimeout;
	function nextFrame() {
		var now = (new Date()).getTime()/1000;
		var framesCounted = 0;
		while ( framesCounted < frameJumpMax && nextFrameIdx < tty_data.length && tty_data[nextFrameIdx].time + time_diff - now < 0 ) {
			record = tty_data[nextFrameIdx++];
			vtview.parseData(record.data);
			framesCounted++;
		}

		vtview.draw();

		if ( nextFrameIdx < tty_data.length )
			nextFrameTimeout = setTimeout(nextFrame, (tty_data[0].time + time_diff - (new Date()).getTime()/1000) * 1000 + accurateTimeInterval);
	}

	setTimeout(function () {
		var freezeData;

		document.getElementById("pause").addEventListener('click', function (evt) {
			if ( nextFrameTimeout ) {
				clearTimeout(nextFrameTimeout);
				nextFrameTimeout = null;
			}
			evt.preventDefault();
		}, false);
		document.getElementById("play").addEventListener('click', function (evt) {
			if ( !nextFrameTimeout )
				go();
			evt.preventDefault();
		}, false);
/*
		document.getElementById("playpause").addEventListener('click', function (evt) {
			evt.preventDefault();
		}, false);
*/
		document.getElementById("freeze").addEventListener('click', function (evt) {
			freezeData = { vtview: vtview.freeze(), nextFrameIdx: nextFrameIdx };
			evt.preventDefault();
		}, false);
		document.getElementById("thaw").addEventListener('click', function (evt) {
			evt.preventDefault();
			var wasplaying = false;
			if ( freezeData ) {
				if ( nextFrameTimeout ) {
					wasplaying = true;
					clearTimeout(nextFrameTimeout);
					nextFrameTimeout = null;
				}

				nextFrameIdx = freezeData.nextFrameIdx;
				vtview.thaw( freezeData.vtview );
				vtview.draw();
				if ( wasplaying )
					go();
			}
		}, false);
		document.getElementById("rewind").addEventListener('click', function (evt) {
			evt.preventDefault();
			var wasplaying = false;
			//if ( freezeData ) {
				if ( nextFrameTimeout ) {
					wasplaying = true;
					clearTimeout(nextFrameTimeout);
					nextFrameTimeout = null;
				}

				nextFrameIdx = initialState.nextFrameIdx;
				vtview.thaw( initialState.vtview );
				vtview.draw();
				if ( wasplaying )
					go();
			//}
		}, false);
	}, 100);
</script>

	<canvas id="canvas" width=80 height=24>No canvas support!</canvas>
	<form>
		<input type="submit" value="Play" id="play">
		<input type="submit" value="Pause" id="pause">
		<!-- <input type="submit" value="PlayPause" id="playpause"> -->
		<input type="submit" value="Freeze" id="freeze">
		<input type="submit" value="Thaw" id="thaw">
		<input type="submit" value="Rewind" id="rewind">
	</form>
EOM;

	return $content;
}

